package com.octacoregames.flappyunicorn;


import android.app.Activity;
import android.content.Context;
import android.content.res.AssetManager;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.os.Bundle;
import android.os.Looper;
import android.util.AttributeSet;
import android.view.Gravity;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.SurfaceHolder;
import android.view.SurfaceView;
import android.view.View;
import android.view.ViewGroup.LayoutParams;

import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;

import com.octacoregames.flappyunicorn.SoundManager;

import java.io.*;
import java.util.*;

import com.inneractive.api.ads.InneractiveAd;
import com.inneractive.api.ads.InneractiveAdListener;
import com.heyzap.sdk.ads.HeyzapAds;
import com.heyzap.sdk.ads.InterstitialAd;

public class GameScreen extends SurfaceView implements SurfaceHolder.Callback, Runnable {
	
	//-------------------------------- Android-specific ----------------------------------
	
	//Image ioff; //double-buffering
	//Multi-Touch
	private Vector<Touch> touches = new Vector<Touch>();
    /** Handle to the application context, used to e.g. fetch Drawables. */
    private Context mContext;
    
    /** Handle to the surface manager object we interact with */
    private SurfaceHolder mSurfaceHolder;

    private Thread thread = null;

    private Graphics g = Graphics.instance();
    private boolean running = false;
    private boolean firstTime = true;
    private boolean paused = false;

	public GameScreen(Context context, AttributeSet attrs) {
        super(context, attrs);
        // register our interest in hearing about changes to our surface
        mSurfaceHolder = getHolder();
        mSurfaceHolder.addCallback(this);
        mContext = context;
        HeyzapAds.start((Activity)mContext);
        setFocusable(true);
    }

    /**
     * Standard override to get key-press events.
     */
    @Override
    public boolean onKeyDown(int keyCode, KeyEvent msg) {
    	keyPressed(keyCode);
    	return super.onKeyDown(keyCode, msg);
    }

    /**
     * Standard override for key-up. We actually care about these, so we can
     * turn off the engine or stop rotating.
     */
    @Override
    public boolean onKeyUp(int keyCode, KeyEvent msg) {
    	keyReleased(keyCode);
    	return super.onKeyUp(keyCode, msg);
    }
    
    @Override
    public boolean onTouchEvent(MotionEvent event) {
        // MotionEvent reports input details from the touch screen
        // and other input controls. In this case, you are only
        // interested in events where the touch position changed.

    	int action = event.getActionMasked();//event.getAction() & MotionEvent.ACTION_MASK;
    	int index = event.getActionIndex();//(event.getAction() & MotionEvent.ACTION_POINTER_INDEX_MASK) >> MotionEvent.ACTION_POINTER_INDEX_SHIFT;    	
        int id = event.getPointerId(index > -1 ? index : 0);
        int x = (int)event.getX();
        int y = (int)event.getY();
        switch (action) {
	    	case MotionEvent.ACTION_POINTER_DOWN:
	    	{
	        	Touch touch = new Touch(id, (int)event.getX(index), (int)event.getY(index), Touch.PRESSED); 
	       		touches.addElement(touch);
	       		return true;
	    	}
	    	case MotionEvent.ACTION_DOWN:
	    	{
	        	Touch touch = new Touch(event.getPointerId(0), (int)event.getX(), (int)event.getY(), Touch.PRESSED); 
	       		touches.addElement(touch);
	    		pointerPressed(x, y);
	    		return true;
	    	}
	        case MotionEvent.ACTION_MOVE:
	        {
	        	Enumeration<Touch> e = touches.elements();
	        	while(e.hasMoreElements()) {
	        		Touch touch = (Touch)e.nextElement();
	        		int idx = event.findPointerIndex(touch.getId());
	        		if(idx > -1) {
	        			touch.setX((int)event.getX(idx));
	        			touch.setY((int)event.getY(idx));
	        		}
	        		touch.setState(Touch.DRAGGED);
	        	}
	    		pointerDragged(x, y);
	        	return true;
	        }
	        case MotionEvent.ACTION_POINTER_UP:
	        {
	        	Enumeration<Touch> e = touches.elements();
	        	while(e.hasMoreElements()) {
	        		Touch touch = (Touch)e.nextElement();
	        		if(touch.getId() == id)	{
	        			touch.setState(Touch.RELEASED);
	        			touches.removeElement(touch);
	        			break;
	        		}
	            }
	        	return true;
	    	}
	        case MotionEvent.ACTION_CANCEL:
	        case MotionEvent.ACTION_UP:
	        {
	        	Enumeration<Touch> e = touches.elements();
	        	while(e.hasMoreElements()) {
	        		Touch touch = (Touch)e.nextElement();
	        		if(touch.getId() == event.getPointerId(0))	{
	        			touch.setState(Touch.RELEASED);
	        			touches.removeElement(touch);
	        			break;
	        		}
	            }
	    		pointerReleased(x, y);
	        	return true;
	    	}
	    }
		super.onTouchEvent(event);
		return true;
    }

    /**
     * Standard window-focus override. Notice focus lost so we can pause on
     * focus lost. e.g. user switches to take a call.
     */
    @Override
    public void onWindowFocusChanged(boolean hasWindowFocus) {
    	if (hasWindowFocus) showNotify();
    	else hideNotify();
    	super.onWindowFocusChanged(hasWindowFocus);
    }

    /* Callback invoked when the surface dimensions change. */
    public void surfaceChanged(SurfaceHolder holder, int format, int width,
            int height) {
        synchronized (mSurfaceHolder) {
        	sizeChanged(width, height);
        }
    }

    /*
     * Callback invoked when the Surface has been created and is ready to be
     * used.
     */
    public void surfaceCreated(SurfaceHolder holder) {
        // start the thread here so that we don't busy-wait in run()
        // waiting for the surface to be created
    	running = true;
    	paused = false;
    	if(thread == null) thread = new Thread(this);
    	try {
    		thread.start();
    	} catch(IllegalThreadStateException ise) {
    		//just to prevent exception when thread was already started
    	}
    }

    /*
     * Callback invoked when the Surface has been destroyed and must no longer
     * be touched. WARNING: after this method returns, the Surface/Canvas must
     * never be touched again!
     */
    public void surfaceDestroyed(SurfaceHolder holder) {
        // we have to tell thread to shut down & wait for it to finish, or else
        // it might touch the Surface after we return and explode
        boolean retry = true;
        paused = running = false;
        while (retry) {
            try {
                thread.join();
                thread = null;
                retry = false;
            } catch (InterruptedException e) {
            }
        }
    }
    
    public void run() {
    	if(firstTime) {
    		init();
        	firstTime = false;
    	}
    	while(running) {
    		while(paused) {
    			try {
    				Thread.sleep(1000);
    			} catch(InterruptedException ie) {
    				ie.printStackTrace();
    			}
    		}
			long delay = System.currentTimeMillis() + 20L;
    		//long beforeTime = System.nanoTime();
			Canvas c = null;
            try {
                c = mSurfaceHolder.lockCanvas(null);                
                synchronized (mSurfaceHolder) {
                	if(c != null) {
                		g.setCanvas(c);
            			paint(g); //no double-buffering
                		/*if(ioff != null) {
                			Graphics goff = ioff.getGraphics();
                			paint(goff);
                            // don't forget to resize the background image
                			//Bitmap bitmap = Bitmap.createScaledBitmap(
                				//	ioff.getBitmap(), getWidth(), getHeight(), true);
                			
                            // Draw the background image. Operations on the Canvas accumulate
                            // so this is like clearing the screen.
                			//c.drawBitmap(bitmap, 0, 0, null);
                			c.drawBitmap(ioff.getBitmap(), 0, 0, null);
                		}*/
                	}
                }
            } catch(Exception e) {
            	//just to prevent IllegalArgumentException when surface is not yet ready
            	//Note: no need for my CM Flare phone but needed for most tablets
            } finally {
                // do this in a finally so that if an exception is thrown
                // during the above, we don't leave the Surface in an
                // inconsistent state
                if (c != null) {
                    mSurfaceHolder.unlockCanvasAndPost(c);
                }
            }
            
            while(delay > System.currentTimeMillis()) Thread.yield();
            /*long sleepTime = 20L-((System.nanoTime()-beforeTime)/1000000L);

            try {
                //actual sleep code
                if(sleepTime>0){
                Thread.sleep(sleepTime);
                }
            } catch (InterruptedException ex) {
            }*/
   	}
    }
    
    /**
     * Restores game state from the indicated Bundle. Typically called when
     * the Activity is being restored after having been previously
     * destroyed.
     *
     * @param savedState Bundle containing the game state
     */
    public synchronized void restoreState(Bundle savedState) {
        synchronized (mSurfaceHolder) {
        	hiscore = savedState.getInt("hiscore");
        }
    }

    /**
     * Dump game state to the provided Bundle. Typically called when the
     * Activity is being suspended.
     *
     * @return Bundle with this view's state
     */
    public Bundle saveState(Bundle map) {
        synchronized (mSurfaceHolder) {
            if (map != null) {
            	map.putInt("hiscore", hiscore);
            }
        }
        return map;
    }
    
	//--------------------------- J2ME wrappers ------------------------	
	// TODO: please move all variable that are always initialized every new game	
	public boolean hasPointerEvents() {
		return true;
	}
	
	public void sizeChanged(int w, int h) {
	}
	
	/** Canvas was hidden from display by interruptions such as Call, SMS, Low Battery etc. */
	public void hideNotify() {
		paused = true;
		if(gameState == STATE_INGAME) setState(STATE_PAUSE);
	}
	
	/** Play bgm if resuming from pause state. Do nothing otherwise */
	public void showNotify() {
		paused = false;
	}

	private void keyDown(int k, boolean isDown) {
		switch(k) {
		case KeyEvent.KEYCODE_DPAD_UP:
			break;
		case KeyEvent.KEYCODE_DPAD_DOWN:
			break;
		case KeyEvent.KEYCODE_DPAD_LEFT:
			break;
		case KeyEvent.KEYCODE_DPAD_RIGHT:
			break;
		case KeyEvent.KEYCODE_SPACE:
			break;
		}
	}
	
	protected void keyPressed(int k) {
		keyDown(k, true);
	}
	
	protected void keyReleased(int k) {
		keyDown(k, false);
	}
	
	public void paint(Graphics g) {
		loopCounter++;
		//long delay = System.currentTimeMillis() + 30L;
		//start demo after 5secs
		if(enableFPS) {
			framesCnt++;
			long currentTime = System.currentTimeMillis();
			if(currentTime - lastTime >= 1000) {
				lastTime = currentTime;
				fps = framesCnt;
				framesCnt = 0;
			}
		}
		
		switch(gameState) {
		case STATE_MAINMENU:
			updateMainMenu(g);
			break;
		case STATE_INGAME:
			updateIngame(g);
			break;
		case STATE_DEMO:
			updateDemo(g);
			break;
		case STATE_PAUSE:
			updatePause(g);
			break;
		case STATE_GAMEOVER:
			updateGameOver(g);
			break;
		}
		if(enableFPS) drawFontString(g, imageLibrary[IMAGE_FONT], "FPS: " + fps, 0, 0, Graphics.LEFT|Graphics.TOP);
		//drawFontString(g, imageLibrary[IMAGE_FONT], "SCR: " + screenWidth +"x"+screenHeight, 0, FONT_HEIGHT << 2, Graphics.LEFT|Graphics.TOP);
	}
	
	// touch screen support, ingame only (no touch screen support in menus)
	boolean touched = false;
	boolean released = false;
	int touchX, touchY;
	protected void pointerPressed(int x, int y) {
		touchX = x * screenWidth / getWidth();
		touchY = y * screenHeight / getHeight();		
		touched = true;
	}
	
	protected void pointerDragged(int x, int y) {
		touchX = x * screenWidth / getWidth();
		touchY = y * screenHeight / getHeight();
	}
	
	protected void pointerReleased(int x, int y) {
		touchX = x * screenWidth / getWidth();
		touchY = y * screenHeight / getHeight();
		touched = false;
		released = true;
	}
	
	//---------------------- common functions -----------------------------
    public void load() {
        
       	try {       		
    		FileInputStream fis = mContext.openFileInput("hiscore");
    		if(fis != null) {
	    		byte[] data = new byte[fis.available()];
	    		fis.read(data);
	    		fis.close();
				ByteArrayInputStream bais = new ByteArrayInputStream(data);
				DataInputStream dis = new DataInputStream(bais);
				hiscore = dis.readInt();
				dis.close();
				bais.close();
    		}
    	} catch(IOException ioe) {
    		ioe.printStackTrace();
    	}
   }
    
    public void save() {
		if(score >= hiscore) {
			hiscore = score;
			MainActivity.getInstance().submitScore(score);
		}
    	try {
        	ByteArrayOutputStream baos = new ByteArrayOutputStream();
        	DataOutputStream dos = new DataOutputStream(baos);
        	dos.writeInt(hiscore);
        	byte[] data = baos.toByteArray();
        	dos.close();
        	baos.close();
    		FileOutputStream fos = mContext.openFileOutput("hiscore", Context.MODE_PRIVATE);
    		fos.write(data);
    		fos.close();
    	} catch(IOException ioe) {
    		ioe.printStackTrace();
    	}
    }

    public String leadingZeroes(int n, int zeroes) {
		String s = "" + n;
		while(s.length() < zeroes) s = "0" + s;
		return s;
	}
	
	public boolean pointInRect(int x, int y, int x1, int y1, int w1, int h1) {
		if(x >= x1 && x <= x1 + w1 && y >= y1 && y <= y1 + h1) {
			return true;
		}
		return false;
	}

	/**
	*	Reference: http://www.gamedev.net/reference/articles/article1275.asp
	*/
	private void customLine(Graphics g, int x1, int y1, int x2, int y2, int size, int lineId) {
		int deltax = Math.abs(x2 - x1);        // The difference between the x's
		int deltay = Math.abs(y2 - y1);        // The difference between the y's
		int x = x1;                       // Start x off at the first pixel
		int y = y1;                       // Start y off at the first pixel
		int xinc1 = 0;
		int xinc2 = 0;
		int yinc1 = 0;
		int yinc2 = 0;
		int den = 0;
		int num = 0;
		int numadd = 0;
		int numpixels = 0;

		if (x2 >= x1)                 // The x-values are increasing
		{
			xinc1 = 1;
			xinc2 = 1;
		}
		else                          // The x-values are decreasing
		{
			xinc1 = -1;
			xinc2 = -1;
		}

		if (y2 >= y1)                 // The y-values are increasing
		{
			yinc1 = 1;
			yinc2 = 1;
		}
		else                          // The y-values are decreasing
		{
			yinc1 = -1;
			yinc2 = -1;
		}

		if (deltax >= deltay)         // There is at least one x-value for every y-value
		{
			xinc1 = 0;                  // Don't change the x when numerator >= denominator
			yinc2 = 0;                  // Don't change the y for every iteration
			den = deltax;
			num = deltax >> 1;
			numadd = deltay;
			numpixels = deltax;         // There are more x-values than y-values
		}
		else                          // There is at least one y-value for every x-value
		{
			xinc2 = 0;                  // Don't change the x for every iteration
			yinc1 = 0;                  // Don't change the y when numerator >= denominator
			den = deltay;
			num = deltay >> 1;
			numadd = deltax;
			numpixels = deltay;         // There are more y-values than x-values
		}

		for (int curpixel = 0; curpixel <= numpixels; curpixel++)
		{
			g.fillRect(x, y, size, size);//size * lineId / MAX_TRAIL_LENGTH, size * lineId / MAX_TRAIL_LENGTH);             // Draw the current pixel
			num += numadd;              // Increase the numerator by the top of the fraction
			if (num >= den)             // Check if numerator >= denominator
			{
				num -= den;               // Calculate the new numerator value
				x += xinc1;               // Change the x as appropriate
				y += yinc1;               // Change the y as appropriate
			}
			x += xinc2;                 // Change the x as appropriate
			y += yinc2;                 // Change the y as appropriate
		}
	}

	public Image getImage(String filename) {
		return getImage(filename, true);
	}
	public Image getImage(String filename, boolean prefix) {
		Image image = null;		
		AssetManager manager = mContext.getAssets();
		try {
			System.out.println("Loading " + filename);
			image = Image.createImage(manager.open((prefix ? ""+BG_W+"/":"")+filename)); //ANDROID: adapts to resolution
		} catch(IOException ioe) {
			System.out.println("Unable to load " + filename);
		}
		return image;
	}
		
	/**
	*	v1.0 Draws a bitmapped font char.
	*/
	public void drawFontChar(Graphics g, Image font, char c, int x, int y) {
		g.setClip(x, y, FONT_WIDTH, FONT_HEIGHT);
		g.drawImage(font, x - ((c - ' ') * FONT_WIDTH), y, Graphics.LEFT|Graphics.TOP);
		g.setClip(0, 0, screenWidth, screenHeight);
	}

	/**
	*	v1.0 Draws a bitmapped font string.
	*/
	public void drawFontString(Graphics g, Image font, String text, int x, int y, int anchor) {
		String s = text.toUpperCase();
		// horizontal alignment
		if((anchor & Graphics.HCENTER) != 0) {
			x -= (FONT_WIDTH * s.length()) >> 1;
		}
		else if((anchor & Graphics.RIGHT) != 0) {
			x -= (FONT_WIDTH * s.length());
		}
		// vertical alignment
		if((anchor & Graphics.VCENTER) != 0) {
			y -= FONT_HEIGHT >> 1;
		}
		else if((anchor & Graphics.BOTTOM) != 0) {
			y -= FONT_HEIGHT;
		}
		for(int i = 0; i < s.length(); i++)
		drawFontChar(g, font, s.charAt(i), x + (i * FONT_WIDTH), y);
	}

	/**
	*	Draws an image out of the source image base on parameters.
	*	@param g the g where the image is to be drawn.
	*	@param src the image source.
	*	@param x_src the x location of the image.
	*	@param y_src the y location of the image.
	*	@param width the width of the image.
	*	@param height the height of the image.
	*	@param x_dest the x destination of the image.
	*	@param y_dest the y destination of the image.
	*	@param anchor the anchor points of the image.
	*/
	public void drawRegion(Graphics g, Image src, int x_src, int y_src, int width, int height, int x_dest, int y_dest, int anchor) {
		g.setClip(x_dest, y_dest, width, height);
		g.drawImage(src, x_dest - x_src, y_dest - y_src, anchor);
		g.setClip(0, 0, screenWidth, screenHeight);
	}
	
	// ref: http://www.java-tips.org
	public Image createThumbnail(Image image, int width, int height) {
		int sourceWidth = image.getWidth();
		int sourceHeight = image.getHeight();

		int thumbWidth = width;
		int thumbHeight = height;

		if (thumbHeight == -1)
			thumbHeight = thumbWidth * sourceHeight / sourceWidth;

		//Need an array (for RGB, with the size of original image)
		int[] rgb = new int[sourceWidth * sourceHeight];
		
		//Get the RGB array of image into "rgb"
		image.getRGB(rgb, 0, sourceWidth, 0, 0, sourceWidth, sourceHeight);
		image = null; System.gc();
		//Call to our function and obtain RGB2
		int[] rgb2 = rescaleArray(rgb, sourceWidth, sourceHeight, thumbWidth, thumbHeight);
		rgb = null; System.gc();
		Image immutableThumb = Image.createRGBImage(rgb2, thumbWidth, thumbHeight, true);
		rgb2 = null; System.gc();
		return immutableThumb;
	}

	private int[] rescaleArray(int[] ini, int x, int y, int x2, int y2) 
	{
		int[] out = new int[x2*y2];

		for (int yy = 0; yy < y2; yy++) 
		{
			int dy = yy * y / y2;
			for (int xx = 0; xx < x2; xx++) 
			{
				int dx = xx * x / x2;
				out[(x2*yy)+xx]=ini[(x*dy)+dx];
			}
		}
		return out;
	}

	/**
	 *	Creates an absolute random number;
	 *	@param max the maximum number to be generated.
	 */
	public int random(int max) {
		return Math.abs(r.nextInt() % max);
	}

	public boolean collision(int x1, int y1, int w1, int h1, int x2, int y2, int w2, int h2) {
		if(x1 + w1 >= x2 && x1 <= x2 + w2 && y1 + h1 >= y2 && y1 <= y2 + h2) return true;
		return false;
	}
	
	//pixel collision
	public byte[] createMask(Image img) {
        // create mask
        byte[] mask;
        int w = img.getWidth();
        int h = img.getHeight();
        int size = w * h;
        int[] argb = new int[size];
        img.getRGB(argb, 0, w, 0, 0, w, h);
        mask = new byte[size];
        for(int i = 0; i < size; i++)
                mask[i] = (byte)((argb[i] & 0xff000000) != 0 ? 1 : 0);
        argb = null;
        System.gc();
        return mask;
    }

    public boolean collision(int x1, int y1, int w1, int h1, int x2, int y2, int w2, int h2, byte[] mask1, byte[] mask2, boolean pixelLevel) {
        if(x1 + w1 >= x2 && x1 <= x2 + w2)
            if(y1 + h1 >= y2 && y1 <= y2 + h2) {
                // is pixel-level?
                if(pixelLevel) {
                    int x, y,		// first mask's offset
                    xx, yy,		// second mask's offset
                    w, h;		// dimension of rect collision area
                    if(x1 < x2) {
                        x = x2 - x1;
                        w = (x1 + w1) - x2;
                        if(w > w2) w = w2;

                        xx = 0;
                    } else {
                        x = 0;
                        w = (x2 + w2) - x1;
                        if(w > w1) w = w1;

                        xx = x1 - x2;
                    }

                    if(y1 < y2) {
                        y = y2 - y1;
                        h = (y1 + h1) - y2;
                        if(h > h2) h = h2;

                        yy = 0;
                    } else {
                        y = 0;
                        h = (y2 + h2) - y1;
                        if(h > h1) h = h1;

                        yy = y1 - y2;
                    }

                    //now check if a pixel on both image of the collision area collides
                    for(int j = 0; j < h; j++)
                        for(int i = 0; i < w; i++)
                            if(getPixel(mask1, w1, x + i, y + j) != 0 && getPixel(mask2, w2, xx + i, yy + j) != 0)
                            return true;
                } else {
                    return true;
                }
            }
        return false;
    }

    public byte getPixel(byte[] mask, int scanLength, int x, int y) {
        return mask[y * scanLength + x];
    }
	
	//---------------------- game-specific --------------------------------	
	final String[] sfx = {"die", "hit", "point", "swooshing", "wing"};
	final int SFX_DIE = 0;
	final int SFX_HIT = 1;
	final int SFX_POINT = 2;
	final int SFX_SWOOSHING = 3;
	final int SFX_WING = 4;
	final int MAX_SFX = 5;
	//Player[] player = new Player[MAX_SFX];

	final int MAX_PIPES = 6;

	//---- scalable vars
	int SCALE = 1; //1
	//fish floats constantly, tap to dive
	int FLAP_SPEED = -15 * SCALE;
	int FALL_SPEED = 1 * SCALE;
	int MAX_FLAP_SPEED = -5 * SCALE;
	int MAX_FALL_SPEED = 10 * SCALE;
	
	int PIPE_W = 48 * SCALE;
	int PIPE_H = 300 * SCALE;
	int PIPE_GAP_X = 160 * SCALE;//160
	int PIPE_GAP_Y = 80 * SCALE;//96
	int PIPE_SPEED = 3 * SCALE;

	int BG_SPEED = 1 * SCALE;
	int BG_W = 480;//320;//420; //TODO: ADJUST THIS (VARIABLE)
	
	int HERO_W = 35 * SCALE;
	int HERO_H = 32 * SCALE;

	int FONT_WIDTH = 16 * SCALE; //16
	int FONT_HEIGHT = 16 * SCALE; //16
	//---- end of scalable vars --------
	
	final int AD_H = 100; //ad height
	
	Random r = null;
	int screenWidth, screenHeight;
	long loopCounter = 0;

	final int IMAGE_BG = 0;
	final int IMAGE_BG1 = IMAGE_BG + 1;
	final int IMAGE_BG2 = IMAGE_BG1 + 1;
	final int IMAGE_HERO = IMAGE_BG2 + 1;
	final int IMAGE_PIPE_DOWN = IMAGE_HERO + 1;
	final int IMAGE_PIPE_UP = IMAGE_PIPE_DOWN + 2;
	final int IMAGE_FONT = IMAGE_PIPE_UP + 2;
	final int IMAGE_PLAY = IMAGE_FONT + 1;
	final int IMAGE_LEADERBOARD = IMAGE_PLAY + 1;
	final int MAX_IMAGES = IMAGE_LEADERBOARD + 1;
	Image[] imageLibrary = new Image[MAX_IMAGES];
	byte[][] mask = null;
	
	int score, hiscore, scoreVisual, scoreboardY, scoreboardVel;
	int ground;
	int pipeId;
	int heroX, heroY, heroSpeed;
	int[] pipeX, pipeY, pipeType;
	boolean alive;	

	//animation
	final int FISH_FRAMES = 12;
	final int FISH_COLORS = 4;
	int currentFrame;
	
	//scrolling background
	int[] bgX;
	boolean enableForeground = false;

	boolean enableFPS = false;

	final int STATE_MAINMENU = 0;
	final int STATE_INGAME = STATE_MAINMENU + 1;
	final int STATE_DEMO = STATE_INGAME + 1;
	final int STATE_PAUSE = STATE_DEMO + 1;
	final int STATE_GAMEOVER = STATE_PAUSE + 1;
	int gameState = STATE_MAINMENU;
	boolean demoMode = false;		
	long demoTimer;
	final long DEMO_DELAY = 15000L;

	int fps = 0;
	int framesCnt = 0;
	long lastTime = 0;
	
	void setState(int state) {
		gameState = state;
	}

	//trail effect
	final int[] rainbow = {
		0xff0000, //red
		0xff8500, //orange
		0xffff00, //yellow
		0x00ff00, //green
		0x0000ff, //blue
		0x4B0082, //indigo
		0xEE82EE //violet
	};
	private final byte MAX_TRAIL_LENGTH = 20;
	private final byte MAX_PROJECTILES = 7;
	private int TRAIL_SIZE = 5;
	private int[][] lineX, lineY;
	private int[] lineIndex;
	private double[] cosTable, sinTable;
	private int trailX, camX;
	private boolean enabledRainbow = false;
	private int rainbowNum = 1;

	//public GameScreen()
	public void init() {
		screenWidth = getWidth();
		screenHeight = getHeight();
		
		//ANDROID: config by resolution
		if(screenWidth < 480) {
			SCALE = 1;
			BG_W = 320; //also supports 240
		} else if(screenWidth < 600/*720*/) {
			SCALE = 2;
			BG_W = 480;
		} else if(screenWidth < 800) {
			SCALE = 3;
			BG_W = 720; //also supports 600
		} else if(screenWidth < 1280) {
			SCALE = 3;
			BG_W = 1080; //also supports 800
		} else if(screenWidth < 1600) {
			SCALE = 4;
			BG_W = 1440; //also supports 1280?
		} else /* greater than 1440 */ {
			SCALE = 6;
			BG_W = 1600;
		}
		if(BG_W < screenWidth) screenWidth = BG_W; //adapt screen to our supported resolution
		FLAP_SPEED *= SCALE;
		FALL_SPEED *= SCALE;
		MAX_FLAP_SPEED *= SCALE;
		MAX_FALL_SPEED *= SCALE;
		PIPE_W *= SCALE;
		PIPE_H *= SCALE;
		PIPE_GAP_X *= SCALE;
		PIPE_GAP_Y *= SCALE;
		PIPE_SPEED *= SCALE;
		HERO_W *= SCALE;
		HERO_H *= SCALE;
		FONT_WIDTH *= SCALE;
		FONT_HEIGHT *= SCALE;
		BG_SPEED *= SCALE;
		//---------------------
		if(enabledRainbow) {
			TRAIL_SIZE *= SCALE;
			lineX = new int[MAX_PROJECTILES][MAX_TRAIL_LENGTH];
			lineY = new int[MAX_PROJECTILES][MAX_TRAIL_LENGTH];
			lineIndex = new int[MAX_PROJECTILES];
			cosTable = new double[360];
			sinTable = new double[360];
			//init trigo table
			for(int i = 0; i < 360; i++) {
				sinTable[i] = Math.sin(Math.toRadians(i));
				cosTable[i] = Math.cos(Math.toRadians(i));
			}
		}
		//ioff = Image.createImage(screenWidth, screenHeight);
		r = new Random(System.currentTimeMillis());

		ground = screenHeight - (FONT_HEIGHT * SCALE);
		currentFrame = 0;

		pipeX = new int[MAX_PIPES];
		pipeY = new int[MAX_PIPES];
		pipeType = new int[MAX_PIPES];
		
		bgX = new int[2];
		for(int i = 0; i < 2; i++) bgX[i] = i * BG_W;
		
		//4 PIPE IMAGES
		mask = new byte[FISH_FRAMES + 4][]; //0 - hero, 1,2 - pipedown, 3,4 - pipeup
		imageLibrary[IMAGE_BG] = getImage("bg0.png");
		imageLibrary[IMAGE_BG1] = getImage("bg1.png");
		//imageLibrary[IMAGE_BG2] = getImage("bg2.png");
		imageLibrary[IMAGE_HERO] = getImage("unicorn.png");
		for(int i = 0; i < FISH_FRAMES; i++)
			mask[i] = createMask(Image.createImage(imageLibrary[IMAGE_HERO], i * HERO_W, 0, HERO_W, HERO_H, 0));
		imageLibrary[IMAGE_PIPE_DOWN] = getImage("pipe_down.png");
		mask[FISH_FRAMES] = createMask(imageLibrary[IMAGE_PIPE_DOWN]);
		imageLibrary[IMAGE_PIPE_DOWN+1] = getImage("pipe_down1.png");
		mask[FISH_FRAMES + 1] = createMask(imageLibrary[IMAGE_PIPE_DOWN+1]);
		imageLibrary[IMAGE_PIPE_UP] = getImage("pipe_up.png");
		mask[FISH_FRAMES + 2] = createMask(imageLibrary[IMAGE_PIPE_UP]);
		imageLibrary[IMAGE_PIPE_UP+1] = getImage("pipe_up1.png");
		mask[FISH_FRAMES + 3] = createMask(imageLibrary[IMAGE_PIPE_UP+1]);
		imageLibrary[IMAGE_FONT] = getImage("font.png");
		imageLibrary[IMAGE_PLAY] = getImage("play.png");
		imageLibrary[IMAGE_LEADERBOARD] = getImage("leaderboard.png");
		

		if(imageLibrary[IMAGE_BG].getHeight() < screenHeight) ground = imageLibrary[IMAGE_BG].getHeight(); 
		initialize();		
		setState(STATE_MAINMENU);
		load();
	}
	
	public void updateMainMenu(Graphics g) {
		if(System.currentTimeMillis() - demoTimer >= DEMO_DELAY) {
			initialize();
			setState(STATE_DEMO);
		}
		if(loopCounter % 3 == 0) {
			currentFrame++;
			currentFrame %= FISH_FRAMES;
		}
		for(int i = 0; i < 2; i++) {
			bgX[i] -= BG_SPEED;
			if(bgX[i] <= -BG_W) bgX[i] += BG_W << 1;
		}
		renderGame(g);
		int offset = AD_H;
		drawFontString(g, imageLibrary[IMAGE_FONT], "Best: " + hiscore, screenWidth >> 1, offset, Graphics.HCENTER|Graphics.TOP);
		offset += FONT_HEIGHT << 2;
		drawFontString(g, imageLibrary[IMAGE_FONT], mContext.getText(R.string.app_name).toString(), screenWidth >> 1, offset, Graphics.HCENTER|Graphics.TOP);
		offset = screenHeight >> 1;
		offset += FONT_HEIGHT << 2;
		drawFontString(g, imageLibrary[IMAGE_FONT], "Tap to flap", screenWidth >> 1, offset, Graphics.HCENTER|Graphics.TOP);
		offset += FONT_HEIGHT;
		drawFontString(g, imageLibrary[IMAGE_FONT], "Avoid obstacles", screenWidth >> 1, offset, Graphics.HCENTER|Graphics.TOP);
		offset += FONT_HEIGHT << 1;
		int playOffset = 0;
		int leaderboardOffset = 0;
		if(touched) {
			if(pointInRect(touchX, touchY, (screenWidth/4) - (imageLibrary[IMAGE_PLAY].getWidth()/2), offset, imageLibrary[IMAGE_PLAY].getWidth(), imageLibrary[IMAGE_PLAY].getHeight())) {
				playOffset = 4 * SCALE;
			} else if(pointInRect(touchX, touchY, (screenWidth*3/4) - (imageLibrary[IMAGE_LEADERBOARD].getWidth()/2), offset, imageLibrary[IMAGE_LEADERBOARD].getWidth(), imageLibrary[IMAGE_LEADERBOARD].getHeight())) {
				leaderboardOffset = 4 * SCALE;
			}
			//touched = false;
		}
		if(released) {
			if(pointInRect(touchX, touchY, (screenWidth/4) - (imageLibrary[IMAGE_PLAY].getWidth()/2), offset, imageLibrary[IMAGE_PLAY].getWidth(), imageLibrary[IMAGE_PLAY].getHeight())) {
				initialize();
				flap();
				setState(STATE_INGAME);
			} else if(pointInRect(touchX, touchY, (screenWidth*3/4) - (imageLibrary[IMAGE_LEADERBOARD].getWidth()/2), offset, imageLibrary[IMAGE_LEADERBOARD].getWidth(), imageLibrary[IMAGE_LEADERBOARD].getHeight())) {
				MainActivity.getInstance().showLeaderboard();
			}
			released = false;
		}
		g.drawImage(imageLibrary[IMAGE_PLAY], screenWidth / 4 + playOffset, offset + playOffset, Graphics.HCENTER|Graphics.TOP);
		g.drawImage(imageLibrary[IMAGE_LEADERBOARD], screenWidth * 3 / 4 + leaderboardOffset, offset + leaderboardOffset, Graphics.HCENTER|Graphics.TOP);
		offset += FONT_HEIGHT;
		drawFontString(g, imageLibrary[IMAGE_FONT], "#Octacore Games", screenWidth >> 1, screenHeight, Graphics.HCENTER|Graphics.BOTTOM);
	}
	
	public void updateIngame(Graphics g) {
		
		//process events
		if(touched) {
			flap();
			touched = false;
		}
		updateGame(g);
		int offset = AD_H;
		drawFontString(g, imageLibrary[IMAGE_FONT], "" + score, screenWidth >> 1, offset, Graphics.HCENTER|Graphics.TOP);
	}
	
	public void updateDemo(Graphics g) {
		if(touched) {
			initialize();
			setState(STATE_MAINMENU);
			touched = false;
		}
		int incomingPipeId = ((pipeId == (MAX_PIPES >> 1) - 1) ? 0 : pipeId + 1) + (MAX_PIPES >> 1);
		if(	heroX < pipeX[incomingPipeId] + PIPE_W &&
			heroY > pipeY[incomingPipeId] - (PIPE_GAP_Y/2)) {
			//heroY = pipeY[incomingPipeId] - (PIPE_GAP_Y >> 1);
			flap();
		}
		updateGame(g);
		if(System.currentTimeMillis() % 2000 > 1000) drawFontString(g, imageLibrary[IMAGE_FONT], "Demo Mode", screenWidth >> 1, AD_H, Graphics.HCENTER|Graphics.BOTTOM);
	}
	
	public void updatePause(Graphics g) {
		if(touched) {
			setState(STATE_INGAME);
			touched = false;
		}
		renderGame(g);
		if(System.currentTimeMillis() % 2000 > 1000) drawFontString(g, imageLibrary[IMAGE_FONT], "Paused", screenWidth >> 1, AD_H, Graphics.HCENTER|Graphics.TOP);
		
	}
	
	public void updateGameOver(Graphics g) {
		int offset = screenHeight >> 1;
		if(System.currentTimeMillis() - demoTimer >= DEMO_DELAY) {
			initialize();
			setState(STATE_DEMO);
		}
		updateGame(g);
		if(scoreboardY == offset && loopCounter % 4 == 0) {
			if(scoreVisual < score) scoreVisual++;
			else if(score > hiscore) hiscore = score;
		}
		if(heroY == ground - HERO_H && scoreboardY > offset) {
			if(scoreboardY == screenHeight) SoundManager.getInstance().playSFX(SoundManager.SFX_SWOOSHING);
			scoreboardVel += PIPE_SPEED;
			scoreboardY -= scoreboardVel;
			if(scoreboardY < offset) scoreboardY = offset;
		}
		drawFontString(g, imageLibrary[IMAGE_FONT], "Best:  " + hiscore + (score >= hiscore && score == scoreVisual ? " New" : ""), (screenWidth >> 1) - (FONT_WIDTH * 6), scoreboardY, Graphics.LEFT|Graphics.TOP);
		drawFontString(g, imageLibrary[IMAGE_FONT], "Score: " + scoreVisual, (screenWidth >> 1) - (FONT_WIDTH * 6), scoreboardY + (FONT_HEIGHT << 1), Graphics.LEFT|Graphics.TOP);
		if(System.currentTimeMillis() % 2000 > 1000) drawFontString(g, imageLibrary[IMAGE_FONT], "Game Over!", screenWidth >> 1, AD_H, Graphics.HCENTER|Graphics.TOP);
		
		offset += FONT_HEIGHT * 7; //same position in main menu
		int playOffset = 0;
		int leaderboardOffset = 0;
		if(touched) {
			InterstitialAd.dismiss();
			if(scoreVisual < score) {
				scoreboardY = offset;
				scoreVisual = score;
				if(score > hiscore) hiscore = score;				
			}
			if(pointInRect(touchX, touchY, (screenWidth/4) - (imageLibrary[IMAGE_PLAY].getWidth()/2), offset, imageLibrary[IMAGE_PLAY].getWidth(), imageLibrary[IMAGE_PLAY].getHeight())) {
				playOffset = 4 * SCALE;
			} else if(pointInRect(touchX, touchY, (screenWidth*3/4) - (imageLibrary[IMAGE_LEADERBOARD].getWidth()/2), offset, imageLibrary[IMAGE_LEADERBOARD].getWidth(), imageLibrary[IMAGE_LEADERBOARD].getHeight())) {
				leaderboardOffset = 4 * SCALE;
			}
			//touched = false;
		}
		if(released) {
			if(pointInRect(touchX, touchY, (screenWidth/4) - (imageLibrary[IMAGE_PLAY].getWidth()/2), offset, imageLibrary[IMAGE_PLAY].getWidth(), imageLibrary[IMAGE_PLAY].getHeight())) {
				save();
				initialize();
				//setState(STATE_MAINMENU);
				flap();
				setState(STATE_INGAME);
			} else if(pointInRect(touchX, touchY, (screenWidth*3/4) - (imageLibrary[IMAGE_LEADERBOARD].getWidth()/2), offset, imageLibrary[IMAGE_LEADERBOARD].getWidth(), imageLibrary[IMAGE_LEADERBOARD].getHeight())) {
				MainActivity.getInstance().showLeaderboard();
			}
			released = false;
		}
		g.drawImage(imageLibrary[IMAGE_PLAY], screenWidth / 4 + playOffset, offset + playOffset, Graphics.HCENTER|Graphics.TOP);
		g.drawImage(imageLibrary[IMAGE_LEADERBOARD], screenWidth * 3 / 4 + leaderboardOffset, offset + leaderboardOffset, Graphics.HCENTER|Graphics.TOP);
	}
	
	public void updateGame(Graphics g) {
		boolean flash = false;
		heroY += heroSpeed;
		if(heroY < -HERO_H) {
			heroY = -HERO_H;
			if(alive) {
				die();
				flash = true;
			}
		} else if(heroY >= ground - HERO_H) {
			heroY = ground - HERO_H;
			if(alive) {
				die();
				flash = true;
			}
		} else {
			if(loopCounter % 3 == 0) {
				currentFrame++;
				currentFrame %= FISH_FRAMES;
				heroSpeed += FALL_SPEED;
				if(heroSpeed > MAX_FALL_SPEED) heroSpeed = MAX_FALL_SPEED;
				if(!alive && heroSpeed == MAX_FALL_SPEED - FALL_SPEED) SoundManager.getInstance().playSFX(SFX_DIE);
			}
		}
		if(alive) {
			//update trail
			if(enabledRainbow && rainbowNum > 0) {
				trailX += PIPE_SPEED * 3;
				camX = trailX - (heroX + HERO_W / 4);
			}
			//update pipes
			for(int i = 0; i < MAX_PIPES; i++) {
				if(i < MAX_PIPES >> 1) {
					pipeX[i] -= PIPE_SPEED;
					//renew pipe, only check first half
					if(pipeX[i] < -PIPE_W) {
						pipeX[i] += ((MAX_PIPES >> 1) * PIPE_GAP_X);
						int randPipeH = PIPE_H;
						if(PIPE_H > screenHeight >> 1) randPipeH = screenHeight >> 1;
						pipeY[i] = -(PIPE_H-PIPE_GAP_Y)+random((randPipeH-PIPE_GAP_Y));//-PIPE_H+random(PIPE_H*3/4)+(PIPE_GAP_Y/2);
						pipeType[i] = random(2);
						pipeY[(MAX_PIPES >> 1) + i] = pipeY[i] + PIPE_H + PIPE_GAP_Y;
						pipeType[MAX_PIPES / 2 + i] = random(2);
					}
					pipeX[(MAX_PIPES >> 1) + i] = pipeX[i];
				}
				int pipeMaskId = (i / (MAX_PIPES >> 1)) * 2 + pipeType[i];
				if(collision(heroX, heroY, HERO_W, HERO_H, pipeX[i], pipeY[i], PIPE_W, PIPE_H, mask[currentFrame], mask[FISH_FRAMES+pipeMaskId], true)) {
					die();
					flash = true;
				} else if(collision(heroX, heroY, HERO_W, HERO_H, pipeX[i] + PIPE_W + HERO_W, 0, PIPE_W, screenHeight) && pipeId != i % (MAX_PIPES >> 1)) {
					pipeId = i % (MAX_PIPES >> 1); //should not hit same pipe
					score++;
					/*if(enabledRainbow) {
						if(rainbowNum > 0) {
							rainbowNum--;
							score += 2; //total of 3pts if rainbow is active
						} else if(score % (7 * (score / 7)) == 0 && rainbowNum <= 0) {
							rainbowNum = 7;
						}
					}*/
					SoundManager.getInstance().playSFX(SFX_POINT);
				}
			}
			for(int i = 0; i < 2; i++) {
				bgX[i] -= BG_SPEED;
				if(bgX[i] <= -BG_W) bgX[i] += BG_W << 1;
			}
		}
		
		//handle input
		/*int key = getKeyStates();
		if((key & Canvas.FIRE) != 0) {
			running = false;
		}*/
		renderGame(g);
		
		if(flash) {
			g.setColor(0xffffff);
			g.fillRect(0, 0, screenWidth, screenHeight);
		}		
	}
	
	public void renderGame(Graphics g) {
		//render graphics
		for(int i = 0; i < 2; i++) {
			g.drawImage(imageLibrary[IMAGE_BG + i], bgX[i], 0, Graphics.LEFT|Graphics.TOP);
			//g.drawImage(imageLibrary[IMAGE_BG1], bgX[i], ground, Graphics.LEFT|Graphics.BOTTOM);
		}
		//draw pipes
		for(int i = 0; i < MAX_PIPES; i++) {
			g.drawImage(imageLibrary[IMAGE_PIPE_DOWN + (i / (MAX_PIPES >> 1)) * 2 + pipeType[i]], pipeX[i], pipeY[i], Graphics.LEFT|Graphics.TOP);
		}
		//draw hero
		//horiz-sheet
		int heroAngle = heroSpeed * 90 / MAX_FALL_SPEED;
		if(heroAngle < -10) heroAngle = -10; //clamp
		if(heroAngle < 0) heroAngle += 360; //trim

		//draw trail
		if(enabledRainbow && rainbowNum > 0 && alive) {
			for(int j = 0; j < MAX_PROJECTILES; j++) {
				lineX[j][lineIndex[j]] = trailX; lineY[j][lineIndex[j]++] = heroY + (j * PIPE_SPEED) + (HERO_H/2);
				if(lineIndex[j] >= MAX_TRAIL_LENGTH) {
					lineIndex[j] = MAX_TRAIL_LENGTH - 1;
					for(int k = 0; k < MAX_TRAIL_LENGTH - 1; k++) {
						lineX[j][k] = lineX[j][k + 1];
						lineY[j][k] = lineY[j][k + 1];
					}
				}
				//draw trail
				for(int k = 0; k < lineIndex[j] - 1; k++) {
					if(lineIndex[j] > 1) {
						//g.setColor(255 - (0 * k / lineIndex[j]), 255 - (128 * k / lineIndex[j]), 255 - (255* k / lineIndex[j])); //fading red
						g.setColor(rainbow[j]);
						int x1 = lineX[j][k];
						int y1 = lineY[j][k];
						int x2 = lineX[j][k + 1];
						int y2 = lineY[j][k + 1];
						x1 = (int)(-cosTable[heroAngle]) + (x1+8);
						y1 = (int)(-sinTable[heroAngle]) + (y1+8);
						x2 = (int)(-cosTable[heroAngle]) + (x2+8);
						y2 = (int)(-sinTable[heroAngle]) + (y2+8);
						customLine(g, x1 - camX, y1, x2 - camX, y2, TRAIL_SIZE, k);
					}
				}
			}
		}

		g.getCanvas().save();
		g.getCanvas().rotate(heroAngle, heroX + (HERO_W/2), heroY + (HERO_H/2));
		g.drawRegion(imageLibrary[IMAGE_HERO], currentFrame * HERO_W, 0, HERO_W, HERO_H, 0, heroX, heroY, Graphics.LEFT|Graphics.TOP);
		g.getCanvas().restore();
		//vert-sheet
		//g.drawRegion(imageLibrary[IMAGE_HERO], 0, currentFrame * HERO_H, HERO_W, HERO_H, 0, heroX, heroY, Graphics.LEFT|Graphics.TOP);
		
		//draw ground
		g.setColor(0x4E2E28);
		g.fillRect(0, ground, screenWidth, screenHeight - (screenHeight - ground));
		
		//foreground
		/*if(enableForeground) {
			int tileCols = screenWidth / 80 + 1;
			int tileRows = screenHeight / 80 + 1;
			for(int y = 0; y < tileRows; y++)
				for(int x = 0; x < tileCols; x++)
					g.drawImage(imageLibrary[IMAGE_BG2], x * 80, y * 80, Graphics.LEFT|Graphics.TOP);
		}*/
		
	}


	private void initialize() {
		scoreboardY = screenHeight;
		scoreboardVel = 0;
		scoreVisual = 0;
		score = 0;
		pipeId = -1;
		alive = true;
		
		heroX = screenWidth / 4;
		heroY = screenHeight / 2;
		heroSpeed = 0;
		
		int randPipeH = PIPE_H;
		if(PIPE_H > screenHeight >> 1) randPipeH = screenHeight >> 1;
		for(int i = 0; i < MAX_PIPES / 2; i++) {
			pipeX[i] = (i * PIPE_GAP_X) + (screenWidth << 1);
			pipeY[i] = -(PIPE_H-PIPE_GAP_Y)+random((randPipeH-PIPE_GAP_Y));//-PIPE_H+random(PIPE_H*3/4)+(PIPE_GAP_Y/2);
			pipeType[i] = random(2);
			pipeX[MAX_PIPES / 2 + i] = pipeX[i];
			pipeY[MAX_PIPES / 2 + i] = pipeY[i] + PIPE_H + PIPE_GAP_Y;
			pipeType[MAX_PIPES / 2 + i] = random(2);
		}
		if(enabledRainbow) {
			for(int i = 0; i < MAX_PROJECTILES; i++) {
				for(int j = 0; j < MAX_TRAIL_LENGTH; j++) {
					lineX[i][j] = heroX; lineY[i][j] = heroY + (i * PIPE_SPEED) + (HERO_H/2);
				}
			}
			trailX = heroX + HERO_W / 4;
			//rainbowNum = 0;
		}
		demoTimer = System.currentTimeMillis();
	}
	
	public void flap() {
		heroSpeed += FLAP_SPEED;
		if(heroSpeed < MAX_FLAP_SPEED) heroSpeed = MAX_FLAP_SPEED;
		SoundManager.getInstance().playSFX(SoundManager.getInstance().SFX_WING);
	}

	public void die() {
		/*if(gameState == STATE_DEMO)
			setState(STATE_MAINMENU);
		else*/
			setState(STATE_GAMEOVER);
			InterstitialAd.display((Activity)mContext);
//	        InneractiveAd.displayInterstitialAd((Activity)mContext, null, "Individual_FlappyUnicorn_Android", null);
		alive = false;
		SoundManager.getInstance().playSFX(SoundManager.getInstance().SFX_HIT);
		demoTimer = System.currentTimeMillis();
	}

}
